/***************************************************************************
                          maindialog.cpp  -  description                              
                             -------------------                                         
    begin                : Mon Nov 15 1999                                           
    copyright            : (C) 1999 by Peter Gasparik                         
    email                : pgasparik@yahoo.com                                     
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   * 
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <kfiledialog.h>
#include <kapp.h>
#include <qfileinfo.h>
#include <qcursor.h>
#include <kcursor.h>
#include <qmessagebox.h>

#include "maindialog.h"
#include "configureportdialog.h"
#include "pixmaptextlistboxitem.h"

// response codes from boot loader
#define ARE_YOU_THERE	0x09
#define	GO 						0x10
#define NOP						0x01

#define TOBIN(x)       	(x > 0x39 ? x-0x37 : x-0x30)

extern const char *pixmap_doc[];
extern const char *pixmap_docMot[];
extern const char *pixmap_docInt[];


MainDialog::MainDialog(QWidget *parent, const char *name) :
			QDialog(parent, name, true)
{
	initDialog();
	
	connect(AddButton, SIGNAL(clicked()), this, SLOT(slotAdd()));
	connect(RemoveButton, SIGNAL(clicked()), this, SLOT(slotRemove()));
	connect(LoadButton, SIGNAL(clicked()), this, SLOT(slotLoad()));
	connect(HelpButton, SIGNAL(clicked()), this, SLOT(slotHelp()));
	
	FilesListBox->setMultiSelection(true);
	ProgDlg->hide();
	loading = false;
	
	// read last saved files
	KConfig *config = kapp->getConfig();
	config->setGroup("AppLoad");
 	char *tok = strtok(config->readEntry("Files", "").data(), ";");
 	while (tok)
 	{
 		QString name(tok);
 		addListItem(name);
		tok = strtok(NULL, ";");
 	}
 	enableButtons();
 	FilesListBox->setFocus();
}

MainDialog::~MainDialog()
{
}

void MainDialog::enableControls(bool enable)
{
	if (enable)
	{
		ProgDlg->hide();
		LoadButton->setText("&Load");
	}
	else
	{
		ProgDlg->show();
		ProgDlg->reset();
		LoadButton->setText("&Cancel");
	}
	AddButton->setEnabled(enable);
	RemoveButton->setEnabled(enable);
	FilesListBox->setEnabled(enable);
	MainMenu->setEnabled(enable);
}

void MainDialog::writeToConfig()
{
	QString Files;
	for (uint i=0; i < FilesListBox->count(); i++)
	{
		if (i > 0)
			Files += ";";
			
		Files += FilesListBox->text(i);
	}

	KConfig *config = kapp->getConfig();
	config->setGroup("AppLoad");
	config->writeEntry("Files", Files.data());
	config->sync();
}

void MainDialog::enableButtons()
{
	bool enable = FilesListBox->count() > 0;
	RemoveButton->setEnabled(enable);
 	LoadButton->setEnabled(enable);
}

void MainDialog::addListItem(QString &fileName)
{
	// generic pixmap
 	PixmapTextListBoxItem *listItem;
	
	int pos = fileName.findRev(".s19", -1, false);
	if (pos != -1 && (pos + 4) == (int)fileName.length())
		listItem = new PixmapTextListBoxItem(fileName.data(), QPixmap(pixmap_docMot));
	else
	{
		pos = fileName.findRev(".ihx", -1, false);
		if (pos == -1)
			pos = fileName.findRev(".hex", -1, false);
		if (pos != -1 && (pos + 4) == (int)fileName.length())
			listItem = new PixmapTextListBoxItem(fileName.data(), QPixmap(pixmap_docInt));
  	else
 			listItem = new PixmapTextListBoxItem(fileName.data(), QPixmap(pixmap_doc));
	}
	FilesListBox->insertItem(listItem);
	FilesListBox->setSelected(FilesListBox->count()-1, true);
}

void MainDialog::slotAdd()
{
	fname = KFileDialog::getOpenFileName(fname, "*.s19|Motorola S record (*.s19)\n*.ihx *.hex|Intel HEX record (*.ihx *.hex)\n*.*|All Files (*.*)", this);
	if (fname.length() > 0)	
	{
		addListItem(fname);
		
		enableButtons();     	
		writeToConfig();
	}
}

void MainDialog::slotRemove()
{
	bool removed;
	do
	{
		removed = false;
   	for (uint i=0; i < FilesListBox->count(); i++)
   	{
   		if (FilesListBox->isSelected(i))
   		{
   			FilesListBox->removeItem(i);
   			removed = true;
   			break;
   		}
   	}
 	} while (removed);
 	
 	writeToConfig();
 	enableButtons();
}

int MainDialog::getBaud(const char *Key)
{
	KConfig *config = kapp->getConfig();
	config->setGroup("Port");
	
	int baud;
	switch (config->readNumEntry(Key, 0))
	{
		case 1200: 	baud = B1200; break;
		case 2400: 	baud = B2400; break;
		case 4800: 	baud = B4800; break;
		case 9600: 	baud = B9600; break;
		case 19200: baud = B19200; break;
		case 38400: baud = B38400; break;
		default:	baud = B9600; break;
	}
	
	return baud;
}

void MainDialog::slotLoad()
{
	if (loading)
	{
		if (QMessageBox::warning(this, "Confirmation", "Do you wish to cancel application loading?", QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes)
		{
			wasCancelled = true;
			LoadButton->setText("&Load");
		}
		return;
	}
	
	KConfig *config = kapp->getConfig();
	config->setGroup("Port");
	QString serPort(config->readEntry("Port"));
	if (serPort == NO_PORT)
	{
		QMessageBox::warning(this, "Warning", "Serial port hasn't been configured yet.\nUse File|Configure Port menu option.");
		return;		
	}
	if (FilesListBox->count() == 0)
	{
		QMessageBox::warning(this, "Warning", "No files specified in the files list.\nUse Add button to include some files.");
		return;
	}
	int selCount = 0;
	for (int i=0; i < (int)FilesListBox->count(); i++)
	{
		if (FilesListBox->isSelected(i))
			selCount++;
	}
	if (selCount == 0)
	{
		QMessageBox::warning(this, "Warning", "No files have been selected for upload.");
		return;		
	}
	
	QString text;
	text.sprintf("You are about to load %d %s.\n\nPut target device into boot mode and then\nselect Ok button.",
				 selCount, selCount == 1 ? "file" : "files");
	int resp = QMessageBox::information(this, "Information", text.data(), QMessageBox::Ok|QMessageBox::Default, QMessageBox::Cancel|QMessageBox::Escape);
	if (resp == QMessageBox::Cancel)
		return;

	SerialPort comm;
	bool failed = false;
	wasCancelled = false;
	loading = true;

	enableControls(false);	
	MemoEdit->insertLine("Booting ...");
	MemoEdit->setCursorPosition(MemoEdit->numLines(), 0);
	if (loadBootLoader(comm))
	{
		text = MemoEdit->textLine(MemoEdit->numLines()-1);
		text += " Ok";
		MemoEdit->removeLine(MemoEdit->numLines()-1);
		MemoEdit->insertLine(text.data());

		for (int i=0, count=0; i < (int)FilesListBox->count(); i++)
		{
			if (!FilesListBox->isSelected(i))
				continue;

			count++;
			QFileInfo fInfo(FilesListBox->text(i));
			text.sprintf("Loading %s ...", fInfo.fileName().data());
			MemoEdit->insertLine(text.data());
			MemoEdit->setCursorPosition(MemoEdit->numLines(), 0);
			ProgDlg->reset();

			int len = loadApp(FilesListBox->text(i), comm);
			if (wasCancelled)
				break;
				
			text = MemoEdit->textLine(MemoEdit->numLines()-1);
			if (len > 0)
			{
				QString s;
				s.sprintf(" %d bytes loaded Ok", len);
				text += s;
			}
			else
			{
				text += " Failed";
				failed = true;
			}
			MemoEdit->removeLine(MemoEdit->numLines()-1);
			MemoEdit->insertLine(text.data());
		}
	}
	else if (!wasCancelled)
	{
		failed = true;
		text = MemoEdit->textLine(MemoEdit->numLines()-1);
		text += "failed";
		MemoEdit->removeLine(MemoEdit->numLines()-1);
		MemoEdit->insertLine(text.data());
	}
	if (wasCancelled)
		MemoEdit->insertLine("Loading cancelled by user.");
	if (!failed && !wasCancelled)
	{
		text.sprintf("Selected %s loaded, you can restart target now.", (selCount > 1) ? "files" : "file");
		MemoEdit->insertLine(text.data());
	}
	MemoEdit->insertLine("");
	MemoEdit->setCursorPosition(MemoEdit->numLines(), 0);
	enableControls(true);
	loading = false;
}
	

bool MainDialog::loadBootLoader(SerialPort &comm)
{
    QString text;

	// Read the program bLoader into memory
	KConfig *config = kapp->getConfig();
	config->setGroup("Port");
	QString serPort(config->readEntry("Port"));
	QString loader(config->readEntry("BootLoader"));
	FILE *In = fopen(loader.data(), "rt");
	if (In == NULL)
	{
		text.sprintf("Unable to open boot loader file:\n%s", loader.data());
		QMessageBox::warning(this, "Warning", text.data());
		return false;
	}

	byte dataBuff[513];
	int loaderLen = 0;

	memset(dataBuff, NOP, sizeof(dataBuff));
	dataBuff[0] = 0xFF;		// synchro byte
	while(!feof(In))
	{
		char line[256];
		fgets(line, sizeof(line), In);
		if (line[0] != ':' && line[0] != 'S')
		{
			text.sprintf("Unexpected start of line character: %c\nInput files are expected in Intel HEX or Motorala S format.", line[0]);
			QMessageBox::warning(this, "Warning", text.data());
			fclose(In);
			return false;
		}
		if (line[0] == 'S' && line[1] != '1')
			continue;

		int offs = line[0] == ':' ? 1 : 2;
		int len = getByte(line+offs);
		if (line[0] == 'S')
			 len -= 3;
		offs += 2;
		int addr = getWord(line+offs);
		offs += 4;
		if (line[0] == ':')
		{
			int type = getByte(line+offs);
			offs += 2;
			if (type == 1) break;
			if (type != 0) continue;
		}
		if (addr+len > 512)
		{
			QMessageBox::warning(this, "Warning", "Boot loader length is greater than 512 Bytes.");
			fclose(In);
			return false;
		}
		loaderLen += len;
		for (int i=0; i < len; i++, offs += 2)
			dataBuff[addr+i+1] = getByte(line+offs);
	}
	fclose(In);

	// set loader length to either 257 (A series) or 513 (E series)
	loaderLen = (loaderLen > 256) ? sizeof(dataBuff) : 257;

	// Load the boot loader
	int baud = getBaud("BootBaud");
	comm.openPort(serPort.data(), baud);
	if (!comm.portOk())
	{
		text.sprintf("Unable to initialise port:\n%s at %d baud.", serPort.data(), baud);
		QMessageBox::warning(this, "Warning", text.data());
		return false;
	}

	ProgDlg->setTotalSteps(loaderLen);

	// transfer Boot Loader to the HC11
	for (int i=0; i < loaderLen; i++)
	{
		ProgDlg->setProgress(i);
		kapp->processEvents();
		if (wasCancelled)
			return false;
		comm.send(dataBuff[i]);
		comm.waitTxEmpty();
	}
	delay(100);

	// ignore echoed chars
	int avail = comm.avail();
	memset(dataBuff, 0, sizeof(dataBuff));
	comm.receive(dataBuff, avail);

	QApplication::restoreOverrideCursor();
	
	baud = getBaud("LoadBaud");
	comm.openPort(serPort.data(), baud);
	if (!comm.portOk())
	{
		text.sprintf("Unable to initialise port:\n%s at %d baud.", serPort.data(), baud);
		QMessageBox::warning(this, "Warning", text.data());
		return false;
	}
	
	// tell bootloader to start running
	byte go = ARE_YOU_THERE;
	comm.send(go);
	comm.waitTxEmpty();
	delay(100);
	// wait for confirmation	
	go = 0;
	if (!comm.receive(go))
	{
		QMessageBox::warning(this, "Warning", "Timed out waiting for boot loader to start.");
		return false;
	}
	if (go != GO)
	{
		text.sprintf("Boot Loader loading failed.\nExpected response: 0x%02X, Received: 0x%02X", GO, go);
		QMessageBox::warning(this, "Warning", text.data());
		return false;
	}

	return true;
}

int MainDialog::loadApp(const char *fileName, SerialPort &comm)
{
	QString text;

	// Read the program bLoader into memory
	FILE *In = fopen(fileName, "rt");
	if (In == NULL)
	{
		text.sprintf("Unable to open application file:\n%s", fileName);
		QMessageBox::warning(this, "Warning", text.data());
		return false;
	}
	QFileInfo fInfo(fileName);
	ProgDlg->setTotalSteps(fInfo.size());

	int EEstart = 0, EEend = 0;
	if (FileMenu->isItemChecked(EEMenuId))
	{
		KConfig *config = kapp->getConfig();
		config->setGroup("EE");
		sscanf(config->readEntry("EEStart", "B600"), "%4x", &EEstart);
		sscanf(config->readEntry("EEEnd", "B7FF"), "%4x", &EEend);
	}
	bool firstEEbyte = true;

	int totBytes = 0;
	byte dataBuff[128];
	while (true)
	{
		// read one record
		char line[256];
		char *resp = fgets(line, sizeof(line), In);
		if (strlen(line) == 0 || resp == NULL)
			break;

		ProgDlg->setProgress(ftell(In));
		kapp->processEvents();

		// accept only Intel hex and Motorola S1 files
		if (line[0] != ':' && line[0] != 'S')
		{
			text.sprintf("Unexpected start of line character: %c\nInput files are expected in Intel HEX or Motorala S format.", line[0]);
			QMessageBox::warning(this, "Warning", text.data());
			fclose(In);
			return 0;
		}

		if (line[0] == 'S' && line[1] != '1')
			continue;

		int offs = line[0] == ':' ? 1 : 2;
		int len = getByte(line+offs);
		if (line[0] == 'S')
			 len -= 3;
		offs += 2;
		int addr = getWord(line+offs);
		offs += 4;
		if (line[0] == ':')
		{
			int type = getByte(line+offs);
			offs += 2;
			if (type == 1) break;
			if (type != 0) continue;
		}
		for (int i=0; i < len; i++, offs += 2)
			dataBuff[3+i] = getByte(line+offs);

		// Transmit address, length, data and calculate LRC
		byte lrc = 0;
		byte val = (byte)(addr >> 8);
		lrc ^= val;
		dataBuff[0] = val;

		val = (byte)addr;
		lrc ^= val;
		dataBuff[1] = val;

		lrc ^= (byte)len;
		dataBuff[2] = (byte)len;

		for (int i=0; i < len; i++)
			lrc ^= dataBuff[3+i];

		comm.send(dataBuff, 3);
		int currAddr = addr;
		for (int i=0; i < len; i++, currAddr++)
		{
			kapp->processEvents();
			if (wasCancelled)
				return 0;

			comm.send(dataBuff[3+i]);
			if (FileMenu->isItemChecked(EEMenuId) && currAddr >= EEstart && currAddr <= EEend)
			{
				delay(firstEEbyte ? 30 : 15);
				firstEEbyte = false;
			}
		}
		comm.waitTxEmpty();

		// Verify lrc
		byte rcv;
		if (!comm.receive(rcv) || rcv != lrc)
		{
			text.sprintf("lrc doesn't match for data block starting at: 0x%04X\nCalculated: %02X  Received: %02X", addr, lrc, rcv);
			QMessageBox::warning(this, "Warning", text.data());
			fclose(In);
			return 0;
		}
		totBytes += len;
	}
	fclose(In);

	return totBytes;
}

byte MainDialog::getByte(char *data)
{
	byte b = (byte)(TOBIN(data[0]) << 4);
	b |= (byte)TOBIN(data[1]);
	return b;
}

word MainDialog::getWord(char *data)
{
	word w = (word)(getByte(data) << 8);
	w |= (word)getByte(data+2);
	return w;
}

void MainDialog::delay(int timeMs)
{
	QTime time;
	time.start();
	while (time.elapsed() < timeMs)
		kapp->processEvents();
}

void MainDialog::slotHelp()
{
	kapp->invokeHTMLHelp(QString("appload11/index-3.html"), QString("MainDialog"));
}

