/***************************************************************************
                          serialport.cpp  -  description                              
                             -------------------                                         
    begin                : Tue Dec 14 1999                                           
    copyright            : (C) 1999 by Peter Gasparik                         
    email                : pgasparik@yahoo.com                                     
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   * 
 *                                                                         *
 ***************************************************************************/

#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/signal.h>

#include <qstring.h>

#include "serialport.h"


SerialPort::SerialPort()
{
	Handle = -1;
	Initialise();
}

SerialPort::SerialPort(const char *PortName, int Baud, int Format)
{
	Handle = -1;
	Initialise();
	openPort(PortName, Baud, Format);
}

SerialPort::~SerialPort()
{
	if (Handle > 0)
  	close(Handle);
}

void SerialPort::Initialise()
{
	struct sigaction handler;
	
	sigaction(SIGIO, NULL, &handler);
	if (!handler.sa_handler)
		signal(SIGIO, SIG_IGN);
}

bool SerialPort::portExists(const char *PortName)
{
	static struct stat DevStat;
	
	stat(PortName, &DevStat);
	if (S_ISCHR(DevStat.st_mode))
	{
		int fd = open(PortName, O_RDONLY | O_NONBLOCK);
		if (fd > 0)
		{
			close(fd);
			return true;
		}
	}
  return false;
}

bool SerialPort::openPort(const char *PortName, int Baud = B9600, int Format = CS8)
{
	// Close port in case it was already open.
  closePort();

	do
	{
		Handle = open(PortName, O_RDWR | O_NOCTTY | O_NONBLOCK);
	}  while (Handle < 0 && errno == EINTR);
	if (Handle < 0 )
		return false;

	struct termios ttyset;
	ttyset.c_iflag = INPCK;
	ttyset.c_lflag = 0;
	ttyset.c_oflag = 0;
	ttyset.c_cflag = CLOCAL | CREAD | Format;
	ttyset.c_cc[ VMIN ] = 1;
	ttyset.c_cc[ VTIME ] = 0;

	if (cfsetispeed(&ttyset, Baud) < 0 ||
	    cfsetospeed(&ttyset, Baud) < 0 ||
			tcsetattr(Handle, TCSAFLUSH, &ttyset) < 0)
	{
		closePort();
		return false;
	}

	fcntl(Handle, F_SETOWN, getpid());
	fcntl(Handle, F_SETFL, FASYNC);

  return true;
}

void SerialPort::closePort()
{
	if (Handle > 0)
	{
		int result;
  	do
  	{
  		result = close(Handle);
  	}  while (result < 0 && errno == EINTR);
		Handle = -1;
	}
}

int SerialPort::avail()
{
	int result;

	if (ioctl(Handle, FIONREAD, &result) != 0)
		return -1;
		
	return result;
}

void SerialPort::waitTxEmpty()
{
	int result;

	do
	{
		result = tcdrain(Handle);
	} while (result && errno == EINTR);
}

bool SerialPort::receive(byte *Data, int Length, int Timeout)
{
	isOk = false;
	if (Handle < 0)
		return false;
		
	int ret, left, bytes = 0;
	fd_set rfds;
	struct timeval sleep;

	FD_ZERO(&rfds);
	FD_SET(Handle, &rfds);
	sleep.tv_sec = Timeout / 1000;
	sleep.tv_usec = 1000 * (Timeout % 1000);
	left = Length;

	while (bytes < Length)
	{
		if (Timeout > 0)
		{
			do
			{
				ret = select(Handle + 1, &rfds, NULL, NULL, &sleep);
			}  while (ret < 0 && errno == EINTR);
			if (ret == 0 )
				break;
			if (ret < 0)
				return -1;
		}
		ret = read(Handle, Data + bytes, left);
		
		if (ret == 0 )
			break;
			
		if (ret < 0 )
			return -1;
			
		bytes += ret;
		left -= ret;
	}
	return bytes;
	
	return isOk;
}

bool SerialPort::send(byte *Data, int Length)
{
	isOk = false;
	if (Length < 0 || Data == NULL || Handle < 0)
		return isOk;
		
	int result;
 	int total = 0;
	do
	{
		result = write(Handle, Data+total, Length-total);
		if (result > 0)
			total += result;
	}  while ((total < Length) || (result < 0 && errno == EINTR));
	isOk = result >= 0;
	
	return isOk;
}

